from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting


class ULTypeCastValue(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.to_type = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        value = self.get_input(self.value)
        to_type = self.get_input(self.to_type)
        if is_waiting(to_type, value):
            return STATUS_WAITING
        return self.typecast_value(value, to_type)

    def evaluate(self):
        self._set_ready()

    def typecast_value(self, value, t):
        if t == 'int':
            return int(value)
        elif t == 'bool':
            return bool(value)
        elif t == 'str':
            return str(value)
        elif t == 'float':
            return float(value)
        return value



from uplogic.nodes import ULConditionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting
from bge import logic


class ULMouseReleased(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.pulse = False
        self.mouse_button_code = None
        self.network = None
        self.OUT = ULOutSocket(self, self.get_changed)

    def get_changed(self):
        socket = self.get_output('changed')
        if socket is None:
            mouse_button = self.get_input(self.mouse_button_code)
            if is_waiting(mouse_button):
                return STATUS_WAITING
            mstat = logic.mouse.inputs[mouse_button]
            if self.pulse:
                return self.set_output(
                    'changed',
                    (
                        mstat.released or
                        mstat.inactive
                    )
                )
            else:
                return self.set_output(
                    'changed',
                    (mstat.released)
                )
        return socket

    def setup(self, network):
        self.network = network

    def evaluate(self):
        self._set_ready()



from mathutils import Euler
from mathutils import Matrix
from mathutils import Vector
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode


class ULEulerToMatrix(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.input_e = None
        self.matrix = Matrix()
        self.OUT = ULOutSocket(self, self.get_matrix)

    def get_matrix(self):
        socket = self.get_output('matrix')
        if socket is None:
            vec = self.get_input(self.input_e)
            if isinstance(vec, Vector):
                vec = Euler((vec.x, vec.y, vec.z), 'XYZ')
            return self.set_output('matrix', vec.to_matrix())
        return socket

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.shaders import remove_filter
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULRemoveFilter(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.pass_idx = None
        self.done = False
        self.OUT = ULOutSocket(self, self._get_done)

    def _get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        pass_idx = self.get_input(self.pass_idx)
        if is_waiting(pass_idx):
            return
        self._set_ready()
        remove_filter(pass_idx)
        self.done = True



from .shader import ULFilter


glsl = """
uniform sampler2D bgl_RenderedTexture;
in vec4 bgl_TexCoord;

uniform float avgL;
uniform float power;

out vec4 fragColor;

// vec2 texcoord = vec2(bgl_TexCoord[0]).st;

void main(void)
{
    float contrast = avgL;
    float brightness = avgL * power;
    vec4 value = texture(bgl_RenderedTexture, bgl_TexCoord.xy);
    fragColor = (value/contrast) - brightness;
}"""


class AdaptiveToneMapping(ULFilter):

    def __init__(self, power=1.0, avgL=2.0, idx: int = None) -> None:
        self.settings = {'power': float(power), 'avgL': float(avgL)}
        super().__init__(glsl, idx, {'power': self.settings, 'avgL': self.settings})

    @property
    def power(self):
        return self.settings['power']

    @power.setter
    def power(self, val):
        self.settings['power'] = float(val)


class NoXRSessionError(Exception):
    """No XR Session found."""
    def __init__(self, message="No active XR Session found!"):
        self.message = message
        super().__init__(self.message)


class PassIndexOccupiedError(Exception):
    """2D Filter pass index already in-use."""
    def __init__(self, idx=0):
        self.message = f"2D Filter pass index {idx} already in-use!"
        super().__init__(self.message)


class LogicControllerNotSupportedError(Exception):
    """Expression/Python not supported for controller."""
    def __init__(self, message="Expression/Python not supported for controller!"):
        self.message = message
        super().__init__(self.message)


class TypeMismatchError(Exception):
    """Type mismatch."""
    def __init__(self, msg='Type Mismatch!'):
        super().__init__(msg)


from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting
from uplogic.utils import is_invalid


class ULVectorMath(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.op = None
        self.vector = None
        self.vector_2 = None
        self.factor = None
        self.scale = None
        self.vector_3 = None
        self.ior = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        op = self.get_input(self.op)
        vector = self.get_input(self.vector)
        vector_2 = self.get_input(self.vector_2)
        vector_3 = self.get_input(self.vector_3)
        factor = self.get_input(self.factor)
        scale = self.get_input(self.scale)
        if is_waiting(
            op,
            factor
        ):
            return STATUS_WAITING
        if is_invalid(
            vector,
            vector_2
        ):
            return STATUS_WAITING
        calculations = {
            'scale': self.get_scale,
            'length': self.get_length,
            'distance': self.get_distance,
            'dot': self.get_dot,
            'faceforward': self.get_faceforward,
            'refract': self.get_refract,
            'reflect': self.get_reflect,
            'project': self.get_project,
            'cross': self.get_cross,
            'multadd': self.get_multadd,
            'divide': self.get_divide,
            'multiply': self.get_multiply,
            'subtract': self.get_subtract,
            'add': self.get_add,
            'normalize': self.get_normalize,
            'lerp': self.get_lerp,
            'negate': self.get_negate,
        }
        if vector:
            vector = vector.copy()
        if vector_2:
            vector_2 = vector_2.copy()
        if vector_3:
            vector_3 = vector_3.copy()
        return calculations[op](vector, vector_2, vector_3, factor, scale)

    def evaluate(self):
        self._set_ready()

    def get_scale(self, vec, vec2, vec3, fac, scale):
        return vec * scale

    def get_length(self, vec, vec2, vec3, fac, scale):
        return vec.length

    def get_distance(self, vec, vec2, vec3, fac, scale):
        return (vec - vec2).length

    def get_dot(self, vec, vec2, vec3, fac, scale):
        return vec.dot(vec2)

    def get_faceforward(self, vec, vec2, vec3, fac, scale):
        pass

    def get_refract(self, vec, vec2, vec3, fac, scale):
        pass

    def get_reflect(self, vec, vec2, vec3, fac, scale):
        return vec.reflect(vec2)

    def get_project(self, vec, vec2, vec3, fac, scale):
        return vec.project(vec2)

    def get_cross(self, vec, vec2, vec3, fac, scale):
        return vec.cross(vec2)

    def get_multadd(self, vec, vec2, vec3, fac, scale):
        return (vec * vec2) + vec3

    def get_divide(self, vec, vec2, vec3, fac, scale):
        return vec / vec2

    def get_multiply(self, vec, vec2, vec3, fac, scale):
        return vec * vec2

    def get_subtract(self, vec, vec2, vec3, fac, scale):
        return vec - vec2

    def get_add(self, vec, vec2, vec3, fac, scale):
        return vec + vec2

    def get_normalize(self, vec, vec2, vec3, fac, scale):
        return vec.normalized()

    def get_lerp(self, vec, vec2, vec3, fac, scale):
        return vec.lerp(vec2, fac)

    def get_negate(self, vec, vec2, vec3, fac, scale):
        vec.negate()
        return vec



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import not_met
from uplogic.shaders import FXAA
from uplogic.shaders import HBAO
from uplogic.shaders import SSAO
from uplogic.shaders import Vignette
from uplogic.shaders import Brightness
from uplogic.shaders import ChromaticAberration
from uplogic.shaders import Grayscale
from uplogic.shaders import Levels
from uplogic.shaders import Mist


class ULAddFilter(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.pass_idx = None
        self.brightness = None
        self.power = None
        self.density = None
        self.color = None
        self.start = None
        self.end = None
        self.filter = None
        self.filter_type = 'FXAA'
        self.done = False
        self.OUT = ULOutSocket(self, self._get_done)

    def _get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        ftype = self.filter_type
        if not_met(condition) or self.filter:
            if self.filter:
                if ftype == 'VIGNETTE':
                    self.filter.settings['power'] = self.get_input(self.power)
                    self.filter.settings['color'] = self.get_input(self.color)
                elif ftype == 'BRIGHTNESS':
                    self.filter.settings['brightness'] = self.get_input(self.brightness)
                elif ftype == 'CHROMAB':
                    self.filter.settings['power'] = self.get_input(self.power)
                elif ftype == 'GRAYSCALE':
                    self.filter.settings['power'] = self.get_input(self.power)
                elif ftype == 'LEVELS':
                    self.filter.settings['color'] = self.get_input(self.color)
                elif ftype == 'MIST':
                    self.filter.settings['power'] = self.get_input(self.power)
                    self.filter.settings['color'] = self.get_input(self.color)
                    self.filter.settings['start'] = self.get_input(self.start)
                    self.filter.settings['density'] = self.get_input(self.density)
                elif ftype in ['HBAO', 'SSAO']:
                    self.filter.settings['power'] = self.get_input(self.power)
            return
        self._set_ready()

        if ftype == 'FXAA':
            self.filter = FXAA(self.get_input(self.pass_idx))
        elif ftype == 'HBAO':
            self.filter = HBAO(self.get_input(self.power), self.get_input(self.pass_idx))
        elif ftype == 'SSAO':
            self.filter = SSAO(self.get_input(self.power), self.get_input(self.pass_idx))
        elif ftype == 'VIGNETTE':
            self.filter = Vignette(
                self.get_input(self.power),
                self.get_input(self.color),
                self.get_input(self.pass_idx)
            )
        elif ftype == 'BRIGHTNESS':
            self.filter = Brightness(
                self.get_input(self.brightness),
                self.get_input(self.pass_idx)
            )
        elif ftype == 'CHROMAB':
            self.filter = ChromaticAberration(
                self.get_input(self.power),
                self.get_input(self.pass_idx)
            )
        elif ftype == 'GRAYSCALE':
            self.filter = Grayscale(
                self.get_input(self.power),
                self.get_input(self.pass_idx)
            )
        elif ftype == 'LEVELS':
            self.filter = Levels(
                self.get_input(self.color),
                self.get_input(self.pass_idx)
            )
        elif ftype == 'MIST':
            self.filter = Mist(
                self.get_input(self.start),
                self.get_input(self.density),
                self.get_input(self.color),
                self.get_input(self.power),
                self.get_input(self.pass_idx)
            )
        self.done = True



from bge import events
from bge import logic
from bge.types import KX_GameObject as GameObject
from uplogic.audio import ULAudioSystem
from uplogic.data import GlobalDB
from uplogic.nodes import STATUS_READY
from uplogic.nodes import STATUS_WAITING
from uplogic.nodes import ULLogicContainer
from uplogic.utils import debug
from uplogic.utils import load_user_module
from uplogic.utils import make_valid_name
import bpy
# from uplogic import get_mainloop
from bge.types import SCA_PythonKeyboard as Keyboard
from bge.types import SCA_PythonMouse as Mouse
import collections
import time
import Cython


class ULLogicTree(ULLogicContainer):

    def __init__(self):
        from ..input import ULMouse
        ULLogicContainer.__init__(self)
        self._cells: list = []
        self._iter = collections.deque()
        self._lastuid: int = 0
        self._owner: GameObject = None
        self._max_blocking_loop_count: int = 0
        self.keyboard: Keyboard = None
        # self.mouse: Mouse = None
        self.mouse: ULMouse = ULMouse()
        self.keyboard_events = None
        self.active_keyboard_events = None
        self.mouse_events = None
        self.stopped = False
        self.timeline = 0.0
        self._time_then = None
        self.time_per_frame = 0.0
        self._do_remove = False
        self.aud_system_owner = False
        self.init_glob_cats()
        self.audio_system = self.get_aud_system()
        self.sub_networks = []  # a list of networks updated by this network
        self.capslock_pressed = False
        self.evaluated_cells = 0
        scene = logic.getCurrentScene()
        mainloop = scene.get('uplogic.mainloop')
        if mainloop:
            mainloop.logic_tree = self

    def get_aud_system(self):
        aud_sys = GlobalDB.retrieve('uplogic.audio').get('default')
        if not aud_sys:
            self.aud_system_owner = True
            return ULAudioSystem('default')
        return aud_sys

    def init_glob_cats(self):
        if not hasattr(bpy.types.Scene, 'nl_globals_initialized'):
            scene = logic.getCurrentScene()
            cats = getattr(
                bpy.data.scenes[scene.name],
                'nl_global_categories',
                None
            )
            if not cats:
                return

            msg = ''

            dat = {
                'STRING': 'string_val',
                'FLOAT': 'float_val',
                'INTEULR': 'int_val',
                'BOOLEAN': 'bool_val',
                'FILE_PATH': 'filepath_val'
            }

            for c in cats:
                db = GlobalDB.retrieve(c.name)
                msg += f' {c.name},'
                for v in c.content:
                    val = getattr(v, dat.get(v.value_type, 'FLOAT'), 0)
                    db.put(v.name, val, v.persistent)

            if msg:
                debug(f'Globals Initialized:{msg[:-1]}')
            bpy.types.Scene.nl_globals_initialized = True

    def ray_cast(
        self,
        caster_object,
        ray_origin,
        ray_destination,
        property,
        xray,
        distance
    ):
        now = time.time()
        cached_data = caster_object.get("_NL_ray_cast_data")
        if cached_data is not None:
            data_time = cached_data[0]
            data_origin = cached_data[1]
            data_destination = cached_data[2]
            data_property = cached_data[3]
            data_distance = cached_data[4]
            d_time = now - data_time
            # only if we are in the same frame,
            # otherwise scenegraph might have changed
            if d_time < 0.01:
                if (
                    (data_origin == ray_origin) and
                    (data_destination == ray_destination) and
                    (data_property == property) and
                    (data_distance == distance)
                ):
                    return cached_data[5]
            pass
        obj, point, normal = None, None, None
        if property is not None:
            obj, point, normal = caster_object.rayCast(
                ray_destination,
                ray_origin,
                distance,
                property,
                xray=xray
            )
        else:
            obj, point, normal = caster_object.rayCast(
                ray_destination,
                ray_origin,
                distance,
                xray=xray
            )
        cached_data = (
            now,
            ray_origin,
            ray_destination,
            property,
            distance,
            (obj, point, normal)
        )
        caster_object["_NL_ray_cast_data"] = cached_data
        return obj, point, normal

    def set_mouse_position(self, screen_x, screen_y):
        self.mouse.position = (screen_x, screen_y)

    def get_owner(self):
        return self._owner

    def setup(self):
        self.time_per_frame = 0.0
        for cell in self._cells:
            cell.network = self
            cell.setup(self)

    def is_running(self):
        return not self.stopped

    def is_stopped(self):
        return self.stopped

    def stop(self, network=None):
        if self.stopped:
            return
        self._time_then = None
        self.stopped = True
        for cell in self._cells:
            cell.stop(self)

    def _generate_cell_uid(self):
        self._lastuid += 1
        return self._lastuid

    def add_cell(self, cell):
        self._cells.append(cell)
        self._iter.append(cell)
        self._max_blocking_loop_count = len(self._cells) * len(self._cells)
        cell._uid = self._generate_cell_uid()
        return cell

    def evaluate(self):
        now = time.time()
        if self._time_then is None:
            self._time_then = now
        dtime = now - self._time_then
        self._time_then = now
        self.timeline += dtime
        self.time_per_frame = dtime
        if self._owner.invalid:
            print("Network Owner removed from game. Shutting down the network")
            return True
        self.keyboard = logic.keyboard
        self.keyboard_events = self.keyboard.inputs.copy()
        self.active_keyboard_events = self.keyboard.activeInputs.copy()
        caps_lock_event = self.keyboard_events[events.CAPSLOCKKEY]
        if(caps_lock_event.released):
            self.capslock_pressed = not self.capslock_pressed

        # update the cells
        cells = self._iter
        max_loop_count = len(cells)
        loop_index = 0
        done_cells = []
        max_blocking_loop_count = self._max_blocking_loop_count
        while cells:
            if loop_index == max_blocking_loop_count:
                debug(
                    "Network found a blocking condition" +
                    " (due to unconnected or non responsive cell)"
                )
                debug("Stopping network...")
                self.stop()
                return
            cell = cells.popleft()
            if cell in done_cells:
                continue
            else:
                done_cells.append(cell)
            cell.evaluate()
            self.evaluated_cells += 1
            if not cell.has_status(STATUS_READY):
                cells.append(cell)
            loop_index += 1
        done_cells = []
        if(loop_index > max_loop_count):
            debug(
                "Wrong sorting alghorithm!",
                loop_index,
                max_loop_count
            )
        for cell in self._cells:
            cell.reset()
            if cell.has_status(STATUS_WAITING):
                cells.append(cell)
        # pulse subnetworks
        for network in self.sub_networks:
            if network._owner.invalid:
                self.sub_networks.remove(network)
            elif not network.stopped:
                network.evaluate()

    def install_subnetwork(self, owner_object, node_tree_name, initial_status):
        # transform the tree name into a NL module name
        tree_name = make_valid_name(node_tree_name)
        mname = f'nl_{tree_name.lower()}'
        module = load_user_module(mname)
        owner_object[f'NL__{tree_name}'] = initial_status
        tree = module.get_tree(owner_object).network
        owner_object[f'IGNLTree_{tree_name}'] = tree
        self.sub_networks.append(tree)
        tree.stopped = not initial_status
        return tree



from .shader import ULFilter
from mathutils import Vector


glsl = """
uniform sampler2D bgl_RenderedTexture;

in vec4 bgl_TexCoord;
uniform float power;
uniform vec3 color;

out vec4 fragColor;

void main( )
{
	vec2 uv = bgl_TexCoord.xy;
    uv *=  1.0 - uv.yx;
    float vig = uv.x * uv.y * 15;
    vig = pow(vig, power);
    vec4 vcol = vec4(color.x, color.y, color.z, 1);
    vec4 px = texture(bgl_RenderedTexture, bgl_TexCoord.xy);

    fragColor = mix(vcol, px, vig); 
}
"""


class Vignette(ULFilter):

    def __init__(self, power: float = 0.25, color=(0., 0., 0.), idx: int = None) -> None:
        self.settings = {'power': float(power), 'color': Vector(color)}
        super().__init__(glsl, idx, {'power': self.settings, 'color': self.settings})

    @property
    def power(self):
        return self.settings['power']

    @power.setter
    def power(self, val):
        self.settings['power'] = val

    @property
    def color(self):
        return self.settings['color']

    @color.setter
    def color(self, val):
        if not isinstance(val, Vector):
            val = Vector(val)
        self.settings['color'] = val



from .audiosystem import ULAudioSystem  # noqa
from .audiosystem import set_master_volume  # noqa
from .audiosystem import set_lowpass  # noqa
from .audiosystem import set_vr_audio  # noqa
from .audiosystem import get_audio_system  # noqa
from .audiosystem import stop_all_audio  # noqa
from .sound import ULSound2D  # noqa
from .sound import ULSound3D  # noqa
from .sound import ULSpeaker2D  # noqa
from .sound import ULSpeaker3D  # noqa
from .music import ULMusic  # noqa



from uplogic.data import GlobalDB
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULListGlobalValues(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.data_id = None
        self.print_d = None
        self.gv_dict = None
        self.done = False
        self.OUT = ULOutSocket(self.get_done)
        self.VALUE = ULOutSocket(self.get_dict)

    def get_done(self):
        return self.done

    def get_dict(self):
        return self.gv_dict

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        data_id = self.get_input(self.data_id)
        print_d = self.get_input(self.print_d)
        if is_waiting(data_id, print_d):
            return
        self._set_ready()
        db = GlobalDB.retrieve(data_id)
        if print_d:
            print(f'[Logic Nodes] Global category "{data_id}":')
            for e in db.data:
                print('{}\t->\t{}'.format(e, db.data[e]))
            print('END ------------------------------------')
        self.done = True
        self.gv_dict = db.data



from uplogic.nodes import ULParameterNode


class ULKeyCode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.key_code = None

    def evaluate(self):
        self._set_ready()
        key_code = self.get_input(self.key_code)
        self._set_value(key_code)



from mathutils import Vector
from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting
from uplogic.utils import is_invalid


class ULObjectAttribute(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.attribute_name = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        game_object = self.get_input(self.game_object)
        attribute_name = self.get_input(self.attribute_name)
        if is_waiting(game_object, attribute_name):
            return STATUS_WAITING
        if is_invalid(game_object):
            return STATUS_WAITING
        if not hasattr(game_object, attribute_name):
            return STATUS_WAITING
        val = getattr(game_object, attribute_name)
        return val.copy() if isinstance(val, Vector) else val

    def evaluate(self):
        self._set_ready()



from .shader import ULFilter
from .fxaa import FXAA
from .brightness import Brightness
from .vignette import Vignette
from .grayscale import Grayscale
from .adaptivetonemapping import AdaptiveToneMapping
from .hbao import HBAO
from .ssao import SSAO
from .mist import Mist
from .levels import Levels
from .letterbox import Letterbox
from .distort import Distort
from .droplets import Droplets
from .blur import Blur
from .chromaticaberration import ChromaticAberration
from .shader import load_glsl
from .shader import remove_filter
from .shader import toggle_filter
from .shader import set_filter_state



from bge import logic
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import FileLoader


class ULLoadFileContent(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.done = False
        self.updated = False
        self.status = 0.0
        self.datatype = ''
        self.item = ''
        self.loader = None
        self.OUT = ULOutSocket(self, self.get_done)
        self.UPDATED = ULOutSocket(self, self.get_updated)
        self.STATUS = ULOutSocket(self, self.get_status)
        self.DATATYPE = ULOutSocket(self, self.get_datatype)
        self.ITEM = ULOutSocket(self, self.get_item)

    def get_status(self):
        return self.status

    def get_datatype(self):
        return self.datatype

    def get_item(self):
        return self.item

    def get_done(self):
        return self.done

    def get_updated(self):
        return self.updated

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        self._set_ready()
        if condition and self.loader is None:
            self.loader = FileLoader()
        if self.loader:
            self.updated = True
            self.status = self.loader.status
            self.item = self.loader.item
            self.datatype = self.loader.data
            if self.loader.finished:
                self.done = True
                self.loader = None
        else:
            self.updated = False



from mathutils import Vector
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import is_invalid


class ULVectorSplitXYZ(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.input_v = None
        self.output_v = Vector()
        self.OUTX = ULOutSocket(self, self.get_out_x)
        self.OUTY = ULOutSocket(self, self.get_out_y)
        self.OUTZ = ULOutSocket(self, self.get_out_z)

    def get_out_x(self):
        vector = self.get_output('vector')
        if vector is None:
            vector = self.set_output(
                'vector',
                self.get_input(self.input_v)
            )
        return self.output_v.x

    def get_out_y(self):
        return self.output_v.y

    def get_out_z(self):
        return self.output_v.z

    def evaluate(self):
        self._set_ready()
        vec = self.get_input(self.input_v)
        if not is_invalid(vec):
            self.output_v = vec



from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket


class ULStoreValue(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.initialize = True
        self._value = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self._value

    def evaluate(self):
        self._set_ready()
        condition = self.get_input(self.condition)
        if condition or self.initialize:
            self.initialize = False
            self._value = self.get_input(self.value)



from .shader import ULFilter
from bge import logic
from mathutils import Vector


glsl = """
/* Original Code by martins upitis @youtube: https://www.youtube.com/watch?v=UkskiSza4p0
Modified by Iza Zed for uplogic
*/

in vec4 bgl_TexCoord;
#define gl_TexCoord glTexCoord
vec4 glTexCoord[4] = vec4[](bgl_TexCoord,bgl_TexCoord,bgl_TexCoord,bgl_TexCoord);

out vec4 fragColor;
#define gl_FragColor fragColor

uniform sampler2D bgl_RenderedTexture;
uniform float bgl_RenderedTextureWidth;
uniform float bgl_RenderedTextureHeight;

uniform float timer, time, randomtime, blur;
uniform vec3 color;

vec2 texcoord = vec2(gl_TexCoord[0]).st;

const float permTexUnit = 1.0/256.0;		// Perm texture texel-size
const float permTexUnitHalf = 0.5/256.0;	// Half perm texture texel-size

float width = bgl_RenderedTextureWidth;
float height = bgl_RenderedTextureHeight;

//a random texture generator, but you can also use a pre-computed perturbation texture
vec4 rnm(in vec2 tc) 
{
    float noise =  sin(dot(tc ,vec2(12.9898,78.233))) * 43758.5453;

    float noiseR =  fract(noise)*2.0-1.0;
    float noiseG =  fract(noise*1.2154)*2.0-1.0; 
    float noiseB =  fract(noise*1.3453)*2.0-1.0;
    float noiseA =  fract(noise*1.3647)*2.0-1.0;
    
    return vec4(noiseR,noiseG,noiseB,noiseA);
}

vec3 fblur(vec2 uv){
    float samples = 16;
    vec2 resolution = vec2(bgl_RenderedTextureWidth, bgl_RenderedTextureHeight);
    float Pi = 6.28318530718;

    float quality = 3.0;
   
    vec2 radius = blur/resolution.xy;

    vec3 frag = texture(bgl_RenderedTexture, uv).rgb;

    for( float d=0.0; d<Pi; d+=Pi/samples)
    {
		for(float i = 1.0/quality; i<=1.0; i+=1.0/quality)
        {
			frag += texture(bgl_RenderedTexture, uv+vec2(cos(d),sin(d)) * radius * i).rgb;
        }
    }
    frag /= quality * samples - 15.0;
    return frag;
}

float fade(in float t) {
    return t*t*t*(t*(t*6.0-15.0)+10.0);
}

float pnoise3D(in vec3 p)
{
    vec3 pi = permTexUnit*floor(p)+permTexUnitHalf; // Integer part, scaled so +1 moves permTexUnit texel
    // and offset 1/2 texel to sample texel centers
    vec3 pf = fract(p);     // Fractional part for interpolation

    // Noise contributions from (x=0, y=0), z=0 and z=1
    float perm00 = rnm(pi.xy).a ;
    vec3  grad000 = rnm(vec2(perm00, pi.z)).rgb * 4.0 - 1.0;
    float n000 = dot(grad000, pf);
    vec3  grad001 = rnm(vec2(perm00, pi.z + permTexUnit)).rgb * 4.0 - 1.0;
    float n001 = dot(grad001, pf - vec3(0.0, 0.0, 1.0));

    // Noise contributions from (x=0, y=1), z=0 and z=1
    float perm01 = rnm(pi.xy + vec2(0.0, permTexUnit)).a ;
    vec3  grad010 = rnm(vec2(perm01, pi.z)).rgb * 4.0 - 1.0;
    float n010 = dot(grad010, pf - vec3(0.0, 1.0, 0.0));
    vec3  grad011 = rnm(vec2(perm01, pi.z + permTexUnit)).rgb * 4.0 - 1.0;
    float n011 = dot(grad011, pf - vec3(0.0, 1.0, 1.0));

    // Noise contributions from (x=1, y=0), z=0 and z=1
    float perm10 = rnm(pi.xy + vec2(permTexUnit, 0.0)).a ;
    vec3  grad100 = rnm(vec2(perm10, pi.z)).rgb * 4.0 - 1.0;
    float n100 = dot(grad100, pf - vec3(1.0, 0.0, 0.0));
    vec3  grad101 = rnm(vec2(perm10, pi.z + permTexUnit)).rgb * 4.0 - 1.0;
    float n101 = dot(grad101, pf - vec3(1.0, 0.0, 1.0));

    // Noise contributions from (x=1, y=1), z=0 and z=1
    float perm11 = rnm(pi.xy + vec2(permTexUnit, permTexUnit)).a ;
    vec3  grad110 = rnm(vec2(perm11, pi.z)).rgb * 4.0 - 1.0;
    float n110 = dot(grad110, pf - vec3(1.0, 1.0, 0.0));
    vec3  grad111 = rnm(vec2(perm11, pi.z + permTexUnit)).rgb * 4.0 - 1.0;
    float n111 = dot(grad111, pf - vec3(1.0, 1.0, 1.0));

    // Blend contributions along x
    vec4 n_x = mix(vec4(n000, n001, n010, n011), vec4(n100, n101, n110, n111), fade(pf.x));

    // Blend contributions along y
    vec2 n_xy = mix(n_x.xy, n_x.zw, fade(pf.y));

    // Blend contributions along z
    float n_xyz = mix(n_xy.x, n_xy.y, fade(pf.z));

    // We're done, return the final noise value.
    return n_xyz;
}

vec2 coordRot(in vec2 tc, in float angle)
{
    float rotX = ((tc.x*2.0-1.0)*(width/height)*cos(angle)) - ((tc.y*2.0-1.0)*sin(angle));
    float rotY = ((tc.y*2.0-1.0)*cos(angle)) + ((tc.x*2.0-1.0)*(width/height)*sin(angle));
    rotX = ((rotX/(width/height))*0.5+0.5);
    rotY = rotY*0.5+0.5;
    return vec2(rotX,rotY);
}

void main(void)
{
    if (time > 10){
        gl_FragColor = texture(bgl_RenderedTexture, texcoord);
        return;
    }
    float grainsize = 100.0;
    float fade = 12;
    vec2 distortFade = vec2(0.0);
    distortFade.s = clamp(texcoord.s*fade,0.0,1.0);
    distortFade.s -= clamp(1.0-(1.0-texcoord.s)*fade,0.0,1.0);
    distortFade.t = clamp(texcoord.t*fade,0.0,1.0);
    distortFade.t -= clamp(1.0-(1.0-texcoord.t)*fade,0.0,1.0); 

    vec2 rotCoordsR = texcoord;
    
    float dfade = 1.0-pow((1.0-distortFade.s*distortFade.t),2.0);
    float noiz = 0.0;
    float drop = 0.0;

    float timeslow = time*0.02;
    float timefaster = time*0.5;
    
    if (time > 0.0 && time < 4.0)
    {
        noiz += pnoise3D(vec3(texcoord*vec2(width/90.0,height/200.0)+vec2(0.0,timer*0.6),1.0+timer*0.2))*0.25;
        noiz += pnoise3D(vec3(texcoord*vec2(width/1200.0,height/1800.0)+vec2(0.0,timer*0.5),3.0+timer*0.3))*0.75;
    }
    
    if (time > 0.0 && time < 100.0)
    {
        drop += pnoise3D(vec3(texcoord*vec2(width/40.0,height/60.0),randomtime/8.0+timer*0.02))*0.2;
        drop += pnoise3D(vec3(texcoord*vec2(width/80.0,height/200.0),randomtime*2.1+timer*0.03))*0.25;
    }
        
    float dropfade = clamp(time*10.0,0.0,1.0);
    
    float drops = clamp(smoothstep(0.0+timefaster,0.5+timefaster,noiz*0.5+0.5),0.0,1.0);
    float droplet = clamp(smoothstep(0.75+timeslow,1.0+timeslow,drop*0.5+0.5),0.0,1.0);
    
    droplet = pow(clamp(droplet+drops,0.0,1.0),0.1)*3.0;
    float dropletmask = smoothstep(0.77+timeslow,0.79+timeslow,drop*0.5+0.5);

    float mask = smoothstep(0.02+timefaster,0.03+timefaster,noiz*0.5+0.5);
    

    vec2 droplets = vec2(dFdx(texcoord+droplet).r,dFdy(texcoord+droplet).g);		

    vec2 dropcoordR;
    vec2 dropcoordG;
    vec2 dropcoordB;

    droplets = droplets*dfade;
    
    dropcoordR = (texcoord-droplets*1.1);
    dropcoordG = (texcoord-droplets*1.2);	
    dropcoordB = (texcoord-droplets*1.3);	

    vec3 frag = texture(bgl_RenderedTexture, texcoord).rgb;

    vec3 dropletcolor = vec3(0.0);
    if (blur != 0.0){
        dropletcolor.r = fblur(dropcoordR).r * color.x;
        dropletcolor.g = fblur(dropcoordG).g * color.y;
        dropletcolor.b = fblur(dropcoordB).b * color.z;
    } else {
        dropletcolor.r = texture(bgl_RenderedTexture, dropcoordR).r * color.x;
        dropletcolor.g = texture(bgl_RenderedTexture, dropcoordG).g * color.y;
        dropletcolor.b = texture(bgl_RenderedTexture, dropcoordB).b * color.z;
    }
    
    vec3 final = mix(frag, dropletcolor, clamp(dropletmask+mask,0.0,1.0)*dropfade);
    gl_FragColor = vec4(final,1.0);
}"""


class Droplets(ULFilter):

    def __init__(self, color=(1, 1, 1), speed=1.0, blur=0.0, idx: int = None) -> None:
        now = logic.getRealTime()
        self.speed = speed
        self.settings = {
            'blur': float(blur),
            'color': Vector(color),
            'timer': float(now),
            'time': float(100.0),
            'randomtime': float(logic.getRandomFloat())
        }
        self._last_time = now
        self.randomize()
        super().__init__(glsl, idx, {
            'blur': self.settings,
            'color': self.settings,
            'timer': self.settings,
            'time': self.settings,
            'randomtime': self.settings
        })

    def update(self):
        now = logic.getRealTime()
        diff = (now - self._last_time) * self.speed
        self.timer += diff
        self._last_time = now
        self.time += diff
        super().update()

    def randomize(self):
        self.randomtime = (logic.getRandomFloat()*5)*2.0-1.0

    def restart(self):
        self.randomize()
        self.time = 0.0
        self.timer = 0.0

    @property
    def blur(self):
        return self.settings['blur']

    @blur.setter
    def blur(self, val):
        self.settings['blur'] = float(val)

    @property
    def color(self):
        return self.settings['color']

    @color.setter
    def color(self, val):
        self.settings['color'] = Vector(val)

    @property
    def timer(self):
        return self.settings['timer']

    @timer.setter
    def timer(self, val):
        self.settings['timer'] = float(val)

    @property
    def randomtime(self):
        return self.settings['randomtime']

    @randomtime.setter
    def randomtime(self, val):
        self.settings['randomtime'] = float(val)

    @property
    def time(self):
        return self.settings['time']

    @time.setter
    def time(self, val):
        self.settings['time'] = float(val)


from uplogic.nodes import ULConditionNode
from uplogic.utils import is_waiting


class ULMouseScrolled(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.wheel_direction = None

    def evaluate(self):
        wd = self.get_input(self.wheel_direction)
        if is_waiting(wd):
            return
        self._set_ready()
        if wd is None:
            return
        elif wd == 1:  # UP
            self._set_value(self.network.mouse.wheel > 0)
        elif wd == 2:  # DOWN
            self._set_value(self.network.mouse.wheel < 0)
        elif wd == 3:  # UP OR DOWN
            self._set_value(self.network.mouse.wheel != 0)



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid
import random


class ULListIndexRandom(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.condition = None
        self._item = None
        self.items = None
        self.OUT = ULOutSocket(self, self.get_val)

    def get_val(self):
        socket = self.get_output('val')
        if socket is None:
            list_d = self.get_input(self.items)
            if is_invalid(list_d):
                return STATUS_WAITING
            self._item = random.choice(list_d)
            return self.set_output('val', self._item)
        return socket

    def evaluate(self):
        self._set_ready()



'''
'''

from bge import logic
from bge.types import KX_GameObject as GameObject
from random import randint
from random import random
from uplogic.animation import ULActionSystem
from uplogic.animation.actionsystem import get_action_system
from uplogic.events import schedule
import bpy
from uplogic.utils import clamp, debug


PLAY_MODES = {
    'play': logic.KX_ACTION_MODE_PLAY,
    'pingpong': logic.KX_ACTION_MODE_PING_PONG,
    'loop': logic.KX_ACTION_MODE_LOOP
}
"""Available play modes of [`"play"`, `"pingpong"`, `"loop"`]"""


BLEND_MODES = {
    'blend': logic.KX_ACTION_BLEND_BLEND,
    'add': logic.KX_ACTION_BLEND_ADD
}
"""Available blending modes of [`"blend"`, `"add"`]"""


ACTION_STARTED = 'ACTION_STARTED'
ACTION_FINISHED = 'ACTION_FINISHED'


class ULAction():
    '''
    [DEPRECATED]

    Wrapper class for animated actions that provides additional parameters
    and quick access properties.

    :param `game_object`: The `KX_GameObject` on which to play the action.
    :param `action_name`: The name of the action  of `bpy.data.actions`.
    :param `start_frame`: The first frame of the action.
    :param `end_frame`: The last frame of the action.
    :param `layer`: The layer on which to play the action. Leave at -1 for
    auto-selection.
    :param `priority`: The priority with which to play the action (only relevant
    for actions on the same layer).
    :param `blendin`: Use this many frames to "blend into" the animation.
    :param `play_mode`: Mode of playback of [`'play'`, `'loop'`, `'pingpong'`].
    :param `speed`: Playback speed.
    :param `intensity`: "Intensity" of the action; Use this to blend
    animations on different layers together.
    :param `blend_mode`: Mode of blending of [`'blend'`, `'add'`]
    :param `keep`: Whether to keep the animation cached after playback has
    finished. This is useful for setting animation frames regardless of the
    action state.
    '''

    def __init__(
        self,
        game_object: GameObject,
        action_name: str,
        start_frame: int = 0,
        end_frame: int = 250,
        layer: int = -1,
        priority: int = 0,
        blendin: float = 0,
        play_mode: str = 'play',
        speed: float = 1,
        intensity: float = 1,
        blend_mode: str = 'blend',
        keep: bool = False
    ):
        debug('ULAction class will be renamed to "Action" in future releases!')
        self._fps_factor = bpy.context.scene.render.fps / 60
        self._locked = False
        self._speed = speed
        self._frozen_speed = speed
        self.finished = False
        '''Finish state of the animation.'''
        self.keep = keep
        '''Whether to keep or free animation data after playback has finished.'''
        self._intensity = intensity
        self._act_system = get_action_system()
        self.game_object = game_object
        '''The game object the animation is playing on.'''
        self.name = action_name
        '''Name of this action.'''
        self.start_frame = start_frame
        '''Starting Frame of the animation.'''
        self.end_frame = end_frame
        '''End Frame of the animation.'''
        self.priority = priority
        '''Priority of this animation; This is only relevant if multiple
        animations are playing on the same layer.'''
        self.blendin = blendin
        '''The amount of blending frames when starting the animation.'''
        self.layer = layer
        '''The layer the animation is playing on.'''
        self.play_mode = play_mode = PLAY_MODES.get(play_mode, play_mode)
        '''Playback mode of the animation.'''
        self.blend_mode = blend_mode = BLEND_MODES.get(blend_mode, blend_mode)
        '''Blending Mode of the animation.'''
        if layer == -1:
            ULActionSystem.find_free_layer(self)
        elif ULActionSystem.check_layer(self):
            self.finished = True
            return
        layer = self.layer
        same_action = game_object.getActionName(layer) == action_name
        self.on_start()
        if (not same_action and self.is_playing):
            game_object.stopAction(layer)
        if not (self.is_playing or same_action):
            game_object.playAction(
                action_name,
                start_frame,
                end_frame,
                play_mode=play_mode,
                speed=speed,
                layer=layer,
                priority=priority,
                blendin=blendin,
                layer_weight=1-intensity,
                blend_mode=blend_mode
            )
        self.intensity = intensity
        self.speed = speed
        self._act_system.add(self)

    def on_start(self):
        '''Handler for animation playback start.
        '''
        schedule(self, 0, ACTION_STARTED)

    def on_finish(self):
        '''Handler for animation playback finish.
        '''
        schedule(self, 0, ACTION_FINISHED)

    @property
    def is_playing(self) -> bool:
        '''Check if the animation is being played (Read-Only).'''
        if self.game_object.invalid:
            return False
        return self.game_object.isPlayingAction(self.layer)

    @is_playing.setter
    def is_playing(self):
        print('ULAction.is_playing is read-only!')

    @property
    def frame(self) -> float:
        '''Current Frame of the animation.'''
        if self.is_playing:
            return self.game_object.getActionFrame(self.layer)
        return -1

    @frame.setter
    def frame(self, value: float):
        self.game_object.setActionFrame(value, self.layer)

    @property
    def intensity(self) -> float:
        '''Intensity of the animation. Higher layers can be blended over lower
        ones.'''
        return self._intensity

    @intensity.setter
    def intensity(self, value: float):
        if not self.is_playing:
            return
        if not self.is_playing or value == self._intensity:
            return
        self._restart_action()
        self._intensity = clamp(value, 0, 1)

    @property
    def speed(self) -> float:
        '''Playback speed of the animation.'''
        return self._speed

    @speed.setter
    def speed(self, value: float):
        if value < 0.00000000001:
            value = 0.00000000001
        if not self.is_playing or value == self._speed:
            return
        self._speed = value
        self._restart_action()

    def _restart_action(self):
        '''Restart action to use updated values.

        Not intended for manual use.
        '''
        if self._locked:
            return
        self._locked = True
        layer = self.layer
        game_object = self.game_object
        action_name = self.name
        start_frame = self.start_frame
        end_frame = self.end_frame
        play_mode = self.play_mode
        priority = self.priority
        blendin = self.blendin
        intensity = self.intensity
        speed = self.speed * self._fps_factor * logic.getTimeScale()
        blend_mode = self.blend_mode
        frame = self.frame
        reset_frame = (
            start_frame if
            play_mode == logic.KX_ACTION_MODE_LOOP else
            end_frame
        )
        next_frame = (
            frame + speed
            if
            frame + speed <= end_frame
            else
            reset_frame
        )
        game_object.stopAction(layer)
        game_object.playAction(
            action_name,
            start_frame,
            end_frame,
            layer=layer,
            priority=priority,
            blendin=blendin,
            play_mode=play_mode,
            speed=self.speed,
            layer_weight=1 - intensity,
            blend_mode=blend_mode
        )
        game_object.setActionFrame(next_frame, layer)

    def update(self):
        '''This is called each frame.
        '''
        self._locked = False
        game_object = self.game_object
        if game_object.invalid:
            self.remove()
            return
        if not self.intensity:
            return
        layer = self.layer
        start_frame = self.start_frame
        end_frame = self.end_frame
        playing_action = game_object.getActionName(layer)
        playing_frame = game_object.getActionFrame(layer)
        if end_frame < start_frame:
            start_frame, end_frame = end_frame, start_frame
        if (
            (playing_action == self.name) and
            (playing_frame >= start_frame) and
            (playing_frame <= end_frame)
        ):
            if self.play_mode == logic.KX_ACTION_MODE_PLAY:
                if end_frame > start_frame:
                    is_near_end = (playing_frame >= (end_frame))
                else:
                    is_near_end = (playing_frame <= (end_frame))
                if is_near_end and not self.keep:
                    self._act_system.remove(self)

    def remove(self):
        '''Stop and remove this action.
        '''
        self._act_system.remove(self)

    def pause(self):
        '''Pause this action.
        '''
        self._frozen_speed = self.speed
        self.speed = 0

    def unpause(self):
        '''Unpause this action.
        '''
        self.speed = self._frozen_speed

    def stop(self):
        '''Stop playback of this action.
        '''
        self.finished = True
        self.on_finish()
        self.game_object.stopAction(self.layer)

    def restart(self):
        '''Restart this animation with its current parameters.
        '''
        self.finished = False
        self.game_object.stopAction(self.layer)
        self.game_object.playAction(
            self.name,
            self.start_frame,
            self.end_frame,
            layer=self.layer,
            priority=self.priority,
            blendin=self.blendin,
            play_mode=self.play_mode,
            speed=self.speed,
            layer_weight=1 - self.intensity,
            blend_mode=self.blend_mode
        )

    def randomize_frame(self, min: float = -1, max: float = -1):
        '''Randomize the frame of this animation.

        :param `min`: Min range of randomization (Optional).
        :param `max`: Max range of randomization (Optional).
        '''
        if min == -1:
            min = self.start_frame
        if max == -1:
            max = self.end_frame
        frame = randint(min, max)
        self.frame = frame

    def randomize_speed(self, min: float = .9, max: float = 1.1):
        '''Randomize the speed of this animation.

        :param `min`: Min range of randomization (Optional, default 0.9).
        :param `max`: Max range of randomization (Optional, default 1.1).
        '''
        delta = max - min
        self.speed = min + (delta * random())

    def set_frame(self, frame: float):
        '''Set the frame of this action.
        '''
        self.frame = frame

class Action(ULAction):
    '''Wrapper class for animated actions that provides additional parameters
    and quick access properties.

    :param `game_object`: The `KX_GameObject` on which to play the action.
    :param `action_name`: The name of the action  of `bpy.data.actions`.
    :param `start_frame`: The first frame of the action.
    :param `end_frame`: The last frame of the action.
    :param `layer`: The layer on which to play the action. Leave at -1 for
    auto-selection.
    :param `priority`: The priority with which to play the action (only relevant
    for actions on the same layer).
    :param `blendin`: Use this many frames to "blend into" the animation.
    :param `play_mode`: Mode of playback of [`'play'`, `'loop'`, `'pingpong'`].
    :param `speed`: Playback speed.
    :param `intensity`: "Intensity" of the action; Use this to blend
    animations on different layers together.
    :param `blend_mode`: Mode of blending of [`'blend'`, `'add'`]
    :param `keep`: Whether to keep the animation cached after playback has
    finished. This is useful for setting animation frames regardless of the
    action state.'''
    pass


from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import not_met


class ULSetParent(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.child_object = None
        self.parent_object = None
        self.compound = True
        self.ghost = True
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        child_object = self.get_input(self.child_object)
        parent_object = self.get_input(self.parent_object)
        compound = self.get_input(self.compound)
        ghost = self.get_input(self.ghost)
        self._set_ready()
        if is_invalid(child_object, parent_object, compound, ghost):
            return
        if child_object.parent is parent_object:
            return
        child_object.setParent(parent_object, compound, ghost)
        self.done = True



from .shader import ULFilter


glsl = """
/* Original Code by existical @shadertoy: https://www.shadertoy.com/view/Xltfzj
Modified by Iza Zed for UPBGE
*/

uniform sampler2D bgl_RenderedTexture;
uniform float bgl_RenderedTextureWidth;
uniform float bgl_RenderedTextureHeight;
vec2 resolution = vec2(bgl_RenderedTextureWidth, bgl_RenderedTextureHeight);

out vec4 fragColor;
in vec4 bgl_TexCoord;

vec2 texcoord = bgl_TexCoord.xy;

uniform float samples;
uniform float power;

void main()
{
    float Pi = 6.28318530718;

    float quality = 3.0;
   
    vec2 radius = power/resolution.xy;
    
    vec2 uv = texcoord;

    vec4 color = texture(bgl_RenderedTexture, uv);
    
    // Blur calculations
    for( float d=0.0; d<Pi; d+=Pi/samples)
    {
		for(float i=1.0/quality; i<=1.0; i+=1.0/quality)
        {
			color += texture(bgl_RenderedTexture, uv+vec2(cos(d),sin(d)) * radius * i);
        }
    }
    color /= quality * samples - 15.0;
    fragColor = color;
}"""

class Blur(ULFilter):

    def __init__(self, samples=16, power=1.0, idx: int = None) -> None:
       self.settings = {'samples': float(samples), 'power': float(power)}
       super().__init__(glsl, idx, {'samples': self.settings, 'power': self.settings})

    @property
    def samples(self):
        return self.settings['samples']

    @samples.setter
    def samples(self, val):
        self.settings['samples'] = float(val)

    @property
    def power(self):
        return self.settings['power']

    @power.setter
    def power(self, val):
        self.settings['power'] = float(val)



from bge import constraints
from bge import logic
from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met
from uplogic.nodes.logictree import ULLogicTree
import json
import os


class ULSaveGame(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.slot = None
        self.path = ''
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def get_custom_path(self, path):
        if not path.endswith('/') and not path.endswith('json'):
            path = path + '/'
        return path

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        slot = self.get_input(self.slot)
        if is_waiting(slot):
            return
        self._set_ready()
        cust_path = self.get_custom_path(self.path)

        path = (
            logic.expandPath('//Saves/') if self.path == ''
            else cust_path
        )
        os.makedirs(path, exist_ok=True)

        scene = logic.getCurrentScene()
        data = {
            'objects': []
        }

        objs = data['objects']

        for obj in scene.objects:
            if obj.name == '__default__cam__':
                continue
            props = obj.getPropertyNames()
            prop_list = []
            cha = constraints.getCharacter(obj)
            for prop in props:
                if prop != 'NodeTree':
                    if isinstance(obj[prop], ULLogicTree):
                        continue
                    if isinstance(obj[prop], Vector):
                        continue
                    prop_set = {}
                    prop_set['name'] = prop
                    prop_set['value'] = obj[prop]
                    prop_list.append(prop_set)
            
            locloc = obj.localPosition
            locrot = obj.localOrientation.to_euler()
            locsca = obj.localScale
            
            loc = obj.worldPosition
            rot = obj.worldOrientation.to_euler()
            sca = obj.worldScale

            if obj.mass:
                lin_vel = obj.worldLinearVelocity
                ang_vel = obj.worldAngularVelocity
                loclin_vel = obj.localLinearVelocity
                locang_vel = obj.localAngularVelocity

                objs.append(
                    {
                        'name': obj.name,
                        'type': 'rigid_body',
                        'data': {
                            'localPosition': {
                                'x': locloc.x,
                                'y': locloc.y,
                                'z': locloc.z
                            },
                            'localOrientation': {
                                'x': locrot.x,
                                'y': locrot.y,
                                'z': locrot.z
                            },
                            'localScale': {
                                'x': locsca.x,
                                'y': locsca.y,
                                'z': locsca.z
                            },
                            'worldPosition': {
                                'x': loc.x,
                                'y': loc.y,
                                'z': loc.z
                            },
                            'worldOrientation': {
                                'x': rot.x,
                                'y': rot.y,
                                'z': rot.z
                            },
                            'worldLinearVelocity': {
                                'x': lin_vel.x,
                                'y': lin_vel.y,
                                'z': lin_vel.z
                            },
                            'worldAngularVelocity': {
                                'x': ang_vel.x,
                                'y': ang_vel.y,
                                'z': ang_vel.z
                            },
                            'worldScale': {'x': sca.x, 'y': sca.y, 'z': sca.z},
                            'props': prop_list
                        }
                    }
                )
            if cha:
                wDir = cha.walkDirection

                objs.append(
                    {
                        'name': obj.name,
                        'type': 'character',
                        'data': {
                            'localPosition': {
                                'x': locloc.x,
                                'y': locloc.y,
                                'z': locloc.z
                            },
                            'localOrientation': {
                                'x': locrot.x,
                                'y': locrot.y,
                                'z': locrot.z
                            },
                            'localScale': {
                                'x': locsca.x,
                                'y': locsca.y,
                                'z': locsca.z
                            },
                            'worldPosition': {
                                'x': loc.x,
                                'y': loc.y,
                                'z': loc.z
                            },
                            'worldOrientation': {
                                'x': rot.x,
                                'y': rot.y,
                                'z': rot.z
                            },
                            'worldScale': {'x': sca.x, 'y': sca.y, 'z': sca.z},
                            'walkDirection': {
                                'x': wDir.x,
                                'y': wDir.y,
                                'z': wDir.z
                            },
                            'props': prop_list
                        }
                    }
                )
            else:
                objs.append(
                    {
                        'name': obj.name,
                        'type': 'static',
                        'data': {
                            'localPosition': {
                                'x': locloc.x,
                                'y': locloc.y,
                                'z': locloc.z
                            },
                            'localOrientation': {
                                'x': locrot.x,
                                'y': locrot.y,
                                'z': locrot.z
                            },
                            'localScale': {
                                'x': locsca.x,
                                'y': locsca.y,
                                'z': locsca.z
                            },
                            'worldPosition': {
                                'x': loc.x,
                                'y': loc.y,
                                'z': loc.z
                            },
                            'worldOrientation': {
                                'x': rot.x,
                                'y': rot.y,
                                'z': rot.z
                            },
                            'worldScale': {'x': sca.x, 'y': sca.y, 'z': sca.z},
                            'props': prop_list
                        }
                    }
                )
            data['globalDict'] = logic.globalDict

        with open(path + 'save' + str(slot) + ".json", "w") as file:
            json.dump(data, file, indent=2)

        self.done = True



from .shader import ULFilter


glsl = """
uniform sampler2D bgl_RenderedTexture;

in vec4 bgl_TexCoord;
uniform float power;
uniform float factor;

out vec4 fragColor;

void main()
{
    float pow = clamp(1.0 - power, 0.0, 1.0);
	vec4 px =  texture(bgl_RenderedTexture, bgl_TexCoord.xy);
    if (bgl_TexCoord.y > 1.0-factor || bgl_TexCoord.y < factor){
        fragColor = vec4(0.0, 0.0, 0.0, 1.0);
        return;
    }
	fragColor = vec4(
        px.r,
        px.g,
        px.b,
        1.0
    );
}
"""


class Letterbox(ULFilter):

    def __init__(self, power: float = 1.0, factor: float = 0.1, idx: int = None) -> None:
        self.settings = {'power': float(power), 'factor': float(factor)}
        super().__init__(glsl, idx, {'power': self.settings, 'factor': self.settings})

    @property
    def power(self):
        return self.settings['power']

    @power.setter
    def power(self, val):
        self.settings['power'] = float(val)
    
    @property
    def factor(self):
        return self.settings['factor']

    @factor.setter
    def factor(self, val):
        self.settings['factor'] = float(val)



from bge import render
from mathutils import Vector


def draw_line(origin: Vector, target: Vector, color: list = [1, 1, 1, 1]):
    render.drawLine(
        origin,
        target,
        color
    )


def draw_points(points: list, color: list = [1, 1, 1, 1]):
    for i, p in enumerate(points):
        if i < len(points) - 1:
            draw_line(p, points[i+1], color)


def draw_cube(origin: Vector, width: float, color: list = [1, 1, 1, 1], centered=False):
    draw_box(origin, width, width, width, color, centered)


def draw_box(origin: Vector, length: float, width: float, height: float, color: list = [1, 1, 1, 1], centered=False):
    if centered:
        origin = origin.copy() - Vector((length * .5, width * .5, height * .5))

    c1: Vector = origin.copy()
    c2: Vector = origin.copy()
    c3: Vector = origin.copy()
    c4: Vector = origin.copy()
    c5: Vector = origin.copy()
    c6: Vector = origin.copy()
    c7: Vector = origin.copy()

    c1[0] += width

    c2[0] += width
    c2[1] += length

    c3[1] += length

    c4[2] += height

    c5[0] += width
    c5[2] += height

    c6[0] += width
    c6[1] += length
    c6[2] += height

    c7[1] += length
    c7[2] += height

    draw_line(origin, c1, color)
    draw_line(c1, c2, color)
    draw_line(c2, c3, color)
    draw_line(c3, origin, color)

    draw_line(origin, c4, color)
    draw_line(c1, c5, color)
    draw_line(c2, c6, color)
    draw_line(c3, c7, color)

    draw_line(c4, c5, color)
    draw_line(c5, c6, color)
    draw_line(c6, c7, color)
    draw_line(c7, c4, color)



from bge import logic
from uplogic.nodes import ULConditionNode


class ULGamepadActive(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.index = None

    def evaluate(self):
        index = self.get_input(self.index)
        if logic.joysticks[index]:
            joystick = logic.joysticks[index]
        else:
            self._button = False
            self._set_ready()
            return
        self._set_ready()
        axis_active = False
        for x in joystick.axisValues:
            if x < -.1 or x > .1:
                axis_active = True
                break
        self._set_value(
            len(joystick.activeButtons) > 0 or
            axis_active
        )



import json
import os
import bpy


endings = {
    json: '.json'
}


class FileWrapper(dict):

    def __init__(self, filepath, notation=json):
        ending = endings.get(notation, '.json')
        if not filepath.endswith(ending):
            filepath = filepath + ending
        self.filepath = bpy.path.abspath(filepath)
        self.notation = notation
        self.read()

    def read(self):
        f = open(self.filepath, 'r')
        data = self.notation.load(f)
        for dat in data.keys():
            self[dat] = data[dat]
        f.close()

    def write(self):
        path = self.filepath
        notation = self.notation
        if os.path.isfile(path):
            f = open(path, 'w')
            notation.dump(self, f, indent=2)


def load_file(filepath, notation=json):
    if filepath:
        return FileWrapper(filepath, notation)
    else:
        raise FileNotFoundError(f'File {filepath} could not be opened!')


def write_file(path, name, val, notation=json):
    data = None
    ending = endings.get(notation, '.json')
    if not path.endswith(ending):
        path = path + ending
    if os.path.isfile(path):
        f = open(path, 'r')
        data = notation.load(f)
        data[name] = val
        f.close()
        f = open(path, 'w')
        notation.dump(data, f, indent=2)
    else:
        f = open(path, 'w')
        data = {name: val}
        notation.dump(data, f, indent=2)
    f.close()


def read_file(path, name, notation=json):
    data = None
    ending = endings.get(notation, '.json')
    if not path.endswith(ending):
        path = path + ending
    if os.path.isfile(path):
        f = open(path, 'r')
        data = notation.load(f)
        f.close()
        return data.get(name)



from .shader import ULFilter
from mathutils import Vector
from bge import logic, render


glsl = """
in vec4 bgl_TexCoord;
#define gl_TexCoord glTexCoord
vec4 glTexCoord[4] = vec4[](bgl_TexCoord,bgl_TexCoord,bgl_TexCoord,bgl_TexCoord);
out vec4 fragColor;
#define gl_FragColor fragColor

uniform sampler2D bgl_RenderedTexture;
uniform sampler2D bgl_DepthTexture;
vec2 texcoord = vec2(gl_TexCoord[0]).st;
vec2 cancoord = vec2(gl_TexCoord[3]).st;

int model = 3; //mist model
int radial = 1;
uniform float density; //= 1.0;  // mist density
uniform float start; // = 0;  //distance at which mist starts
uniform vec3 color;
uniform float power;

uniform float znear; // = .1; //camera clipping start
uniform float zfar; // = 500; //camera clipping end
uniform float aspect; //camera aspect ratio
uniform float fov; //camera field of view

float linearize(float depth)
{
    return -zfar * znear / (depth * (zfar - znear) - zfar);
}

void main(void)
{    
    float depth = linearize(texture2D(bgl_DepthTexture,texcoord.xy).x);

    if (depth > zfar){
    //Clip the mist to allow for skyboxes
    depth = 0.0;
    }

    if (radial > 0){
        //Mist sphere
        float width = depth * tan(fov / 360 * 3.1415);
        float height = width / aspect;
        float y = (gl_TexCoord[0].st.y - 0.5) * 2.0 * height;
        float x = (gl_TexCoord[0].st.x - 0.5) * 2.0 * width;
        depth = length(vec3(x,y,depth));//pow(depth, 0.3333);
    }    


    vec4 mist = vec4(color.x, color.y, color.z, 1.0);
    float density2 = 1/density;
    float factor = 0.0;
    if (model == 1){
        //Linear Mist
        factor = (depth - start)/(density2*100);
    } else if (model == 2){
        //Quadratic Mist
        factor = max(depth - start, 0.0);
        factor = (factor * factor)/(density2*2000);
    } else if (model == 3){
        //Exponential Mist
        factor = 1.0 - exp(-(depth - start)*density*0.03);
    } else if (model == 4){
        //Exponential Squared Mist
        factor = max(depth - start, 0.0);
        factor = 1.0 - exp(-(factor * factor)*density*density*0.0003);
    }

    vec4 dif = texture(bgl_RenderedTexture, texcoord);

    gl_FragColor = mix(dif, mix(dif, mist, clamp(factor, 0.0, 1.0)), power);

}
"""


class Mist(ULFilter):

    def __init__(self, start=.1, density=0.5, color=(0.5, 0.7, 0.9), power=1.0, idx: int = None) -> None:
        cam = logic.getCurrentScene().active_camera
        self.settings = {
            'start': float(start),
            'density': float(density),
            'color': Vector(color),
            'power': float(power),
            'znear': cam.near,
            'zfar': cam.far,
            'fov': cam.fov,
            'aspect': render.getWindowWidth() / render.getWindowHeight()
        }
        super().__init__(glsl, idx, {
            'start': self.settings,
            'density': self.settings,
            'color': self.settings,
            'power': self.settings,
            'znear': self.settings,
            'zfar': self.settings,
            'fov': self.settings,
            'aspect': self.settings
        })

    def update(self):
        super().update()
        cam = logic.getCurrentScene().active_camera
        self.settings['znear'] = cam.near
        self.settings['zfar'] = cam.far
        self.settings['fov'] = cam.fov
        self.settings['aspect'] = render.getWindowWidth() / render.getWindowHeight()

    @property
    def start(self):
        return self.settings['start']

    @start.setter
    def start(self, val):
        self.settings['start'] = float(val)

    @property
    def power(self):
        return self.settings['power']

    @power.setter
    def power(self, val):
        self.settings['power'] = val

    @property
    def color(self):
        return self.settings['color']

    @color.setter
    def color(self, val):
        if not isinstance(val, Vector):
            val = Vector(val)
        self.settings['color'] = val



from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
import random


class ULRandomVect(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.xyz = None
        self.OUT_A = ULOutSocket(self, self._get_output)

    def _get_output(self):
        socket = self.get_output('result')
        if socket is None:
            xyz = self.get_input(self.xyz)
            if is_waiting(xyz):
                return
            vmin, vmax = -999999999, 999999999
            delta = vmax - vmin
            x = vmin + (delta * random.random()) if xyz['x'] else 0
            y = vmin + (delta * random.random()) if xyz['y'] else 0
            z = vmin + (delta * random.random()) if xyz['z'] else 0
            return self.set_output('result', Vector((x, y, z)))
        return socket

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULConditionNode
from uplogic.utils import is_waiting


class ULKeyReleased(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.pulse = False
        self.key_code = None
        self.network = None

    def setup(self, network):
        self.network = network

    def evaluate(self):
        keycode = self.get_input(self.key_code)
        if is_waiting(keycode):
            return
        self._set_ready()
        keystat = self.network.keyboard_events[keycode]
        if self.pulse:
            self._set_value(
                keystat.released or
                keystat.inactive
            )
        else:
            self._set_value(keystat.released)



from uplogic.events import receive, ULEvent
from uplogic.utils.errors import TypeMismatchError
from bge.types import KX_PythonComponent, KX_GameObject
from bge import logic


def listener(original_class: KX_PythonComponent) -> KX_PythonComponent:
    """`KX_PythonComponent` Class Decorator

    Makes the decorated class listen for events that have the component's
    `object` attribute as ID.
    Executes the component's `on_object` function when an event is detected
    """
    if not issubclass(original_class, KX_PythonComponent):
        raise TypeMismatchError('Decorator only viable for KX_PythonComponent subclasses!')
    orig_init = original_class.__init__

    def __init__(self, obj):
        orig_init(self)
        logic.getCurrentScene().post_draw.append(self._detect_object)

    def _detect_object(self):
        evt: ULEvent = receive(self.object)
        if evt:
            self.on_object(evt)

    original_class._detect_object = _detect_object
    original_class.__init__ = __init__
    return original_class


def state_machine(cls: KX_PythonComponent) -> KX_PythonComponent:
    """`KX_PythonComponent` Class Decorator.

    Automatically adds a `state` property to the `KX_GameObject` of the Component
    that can be accessed by `KX_PythonComponent.state` as well.

    This does not interfere with the built-in `state` attribute of `KX_GameObject`.
    """

    def deco(cls: KX_PythonComponent) -> KX_PythonComponent:
        if not issubclass(cls, KX_PythonComponent):
            raise TypeMismatchError('Decorator only viable for KX_PythonComponent subclasses!')
        def getState(self, attr_name='state'):
            return self.object.get(attr_name)

        def setState(self, value, attr_name='state'):
            if value != self.state:
                self.on_state(value)
            self.object[attr_name] = value

        def set_state(self, state):
            if state != self.state:
                self.on_state(state)
            self.state = state

        def on_state(self, state):
            pass

        prop = property(getState, setState)
        setattr(cls, 'state', prop)
        setattr(cls, 'set_state', set_state)
        if not hasattr(cls, 'on_state'):
            setattr(cls, 'on_state', on_state)
        return cls

    return deco(cls)


def game_props(prop_names) -> KX_PythonComponent:
    """Decorator for `KX_PythonComponent` or `KX_GameObject` classes and subclasses.

    Automatically adds property handlers for this class to use the `game_object[prop]`
    syntax instead of saving values on the instance itself.

    :param `prop_names`: Names of game properties as a list.
    """

    def on_attr(self, val):
        pass

    def deco(cls: KX_PythonComponent or KX_GameObject) -> KX_PythonComponent or KX_GameObject:
        if not (issubclass(cls, KX_PythonComponent) or issubclass(cls, KX_GameObject)):
            raise TypeMismatchError('Decorator only viable for KX_PythonComponent or KX_GameObject subclasses!')
        if not (isinstance(prop_names, list) or isinstance(prop_names, tuple)):
            raise TypeMismatchError('Expected property names as a list or tuple!')
        for game_prop in prop_names:

            def getPropComponent(self, attr_name=game_prop):
                return self.object.get(attr_name)

            def setPropComponent(self, value, attr_name=game_prop):
                getattr(self, f'on_{game_prop}')(value)
                self.object[attr_name] = value

            def getPropObject(self, attr_name=game_prop):
                return self.get(attr_name)

            def setPropObject(self, value, attr_name=game_prop):
                getattr(self, f'on_{game_prop}')(value)
                self[attr_name] = value

            if issubclass(cls, KX_PythonComponent):
                prop = property(getPropComponent, setPropComponent)
            elif issubclass(cls, KX_GameObject):
                prop = property(getPropObject, setPropObject)
            else:
                return

            setattr(cls, game_prop, prop)
            if not hasattr(cls, f'on_{game_prop}'):
                setattr(cls, f'on_{game_prop}', on_attr)
        return cls
    return deco


def bl_attrs(attr_names) -> KX_PythonComponent:
    """Decorator for `KX_PythonComponent` or `KX_GameObject` classes and subclasses.

    Automatically adds attribute handlers for this class to use the
    `game_object.blenderObject[attribute]` syntax instead of saving values on the
    instance itself.

    :param `attr_names`: Names of custom attributes as a list.
    """

    def on_attr(self, val):
        pass

    def deco(cls: KX_PythonComponent or KX_GameObject) -> KX_PythonComponent or KX_GameObject:
        if not (issubclass(cls, KX_PythonComponent) or issubclass(cls, KX_GameObject)):
            raise TypeMismatchError('Decorator only viable for KX_PythonComponent or KX_GameObject subclasses!')
        if not (isinstance(attr_names, list) or isinstance(attr_names, tuple)):
            raise TypeMismatchError('Expected attribute names as a list or tuple!')
        for attr_name in attr_names:

            def getPropComponent(self, attr_name=attr_name):
                return self.object.blenderObject.get(attr_name)

            def setPropComponent(self, value, attr_name=attr_name):
                getattr(self, f'on_{attr_name}')(value)
                self.object.blenderObject[attr_name] = value
                self.object.color = self.object.color

            def getPropObject(self, attr_name=attr_name):
                return self.blenderObject.get(attr_name)

            def setPropObject(self, value, attr_name=attr_name):
                getattr(self, f'on_{attr_name}')(value)
                self.blenderObject[attr_name] = value
                self.object.color = self.object.color

            if issubclass(cls, KX_PythonComponent):
                prop = property(getPropComponent, setPropComponent)
            elif issubclass(cls, KX_GameObject):
                prop = property(getPropObject, setPropObject)
            else:
                return

            setattr(cls, attr_name, prop)
            if not hasattr(cls, f'on_{attr_name}'):
                setattr(cls, f'on_{attr_name}', on_attr)
        return cls
    return deco


def global_dict(prop_names) -> KX_PythonComponent:
    """Decorator for `KX_PythonComponent` or `KX_GameObject` classes and subclasses.

    Automatically adds property handlers for this class to use the `bge.logic.globalDict[key]`
    syntax instead of saving values on the instance itself.

    :param `prop_names`: Keys as a list.
    """
    def deco(cls: KX_PythonComponent or KX_GameObject) -> KX_PythonComponent or KX_GameObject:
        if not (issubclass(cls, KX_PythonComponent) or issubclass(cls, KX_GameObject)):
            raise TypeMismatchError('Decorator only viable for KX_PythonComponent or KX_GameObject subclasses!')
        if not (isinstance(prop_names, list) or isinstance(prop_names, tuple)):
            raise TypeMismatchError('Expected property names as a list or tuple!')
        for game_prop in prop_names:

            def getPropComponent(self, attr_name=game_prop):
                return logic.globalDict.get(attr_name)

            def setPropComponent(self, value, attr_name=game_prop):
                logic.globalDict[attr_name] = value

            prop = property(getPropComponent, setPropComponent)

            setattr(cls, game_prop, prop)
        return cls
    return deco


def scene_props(prop_names):
    """Decorator for `KX_PythonComponent` or `KX_GameObject` classes and subclasses.

    Automatically adds property handlers for this class to use the `game_object.scene[prop]`
    syntax instead of saving values on the instance itself.

    :param `prop_names`: Names of properties as a list.
    """
    def deco(cls: KX_PythonComponent or KX_GameObject) -> KX_PythonComponent or KX_GameObject:
        if not (issubclass(cls, KX_PythonComponent) or issubclass(cls, KX_GameObject)):
            raise TypeMismatchError('Decorator only viable for KX_PythonComponent or KX_GameObject subclasses!')
        if not (isinstance(prop_names, list) or isinstance(prop_names, tuple)):
            raise TypeMismatchError('Expected property names as a list or tuple!')
        for scene_prop in prop_names:

            def getPropComponent(self, attr_name=scene_prop):
                return self.object.scene.get(attr_name)

            def setPropComponent(self, value, attr_name=scene_prop):
                self.object.scene[attr_name] = value

            def getPropObject(self, attr_name=scene_prop):
                return self.scene.get(attr_name)

            def setPropObject(self, value, attr_name=scene_prop):
                self.scene[attr_name] = value

            if issubclass(cls, KX_PythonComponent):
                prop = property(getPropComponent, setPropComponent)
            elif issubclass(cls, KX_GameObject):
                prop = property(getPropObject, setPropObject)
            else:
                return

            setattr(cls, scene_prop, prop)
        return cls
    return deco



from mathutils import Vector
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import is_invalid


class ULVectorXYZ(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.input_x = None
        self.input_y = None
        self.input_z = None
        self.OUTV = ULOutSocket(self, self.get_out_v)

    def get_out_v(self):
        x = self.get_input(self.input_x)
        y = self.get_input(self.input_y)
        z = self.get_input(self.input_z)
        v = Vector((0, 0, 0))
        if not is_invalid(x):
            v.x = x
        if not is_invalid(y):
            v.y = y
        if not is_invalid(y):
            v.z = z
        return v.copy()

    def evaluate(self):
        self._set_ready()



'''TODO: Documentation
'''

from bge import logic
import bpy
import os
from uplogic.utils import debug
from uplogic.utils import unload_nodes


class GlobalDB(object):
    '''TODO: Documentation
    '''
    index: int
    initialized: bool = False

    class LineBuffer(object):
        def __init__(self, buffer=[]):
            self.buffer = buffer
            self.index = 0
            self.size = len(self.buffer)

        def read(self):
            line = self.buffer[self.index]
            self.index += 1
            return line

        def write(self, line):
            self.buffer.append(line + "\n")

        def has_next(self):
            return self.index < self.size

        def flush(self, file):
            with open(file, "a") as f:
                f.writelines(self.buffer)

    class Serializer(object):
        def read(self, line_reader):
            raise NotImplementedError()

        def write(self, value, line_writer):
            raise NotImplementedError()

    serializers = {}
    storage_dir = logic.expandPath("//Globals")
    shared_dbs = {}

    @classmethod
    def retrieve(cls, fname):
        '''TODO: Documentation
        '''
        db = cls.shared_dbs.get(fname)
        if db is None:
            db = GlobalDB(fname)
            cls.shared_dbs[fname] = db
        return db

    @classmethod
    def get_storage_dir(cls):
        '''TODO: Documentation
        '''
        return cls.storage_dir

    @classmethod
    def put_value(cls, key, value, buffer):
        '''TODO: Documentation
        '''
        type_name = str(type(value))
        serializer = cls.serializers.get(type_name)
        if not serializer:
            return False
        buffer.write("PUT")
        buffer.write(key)
        buffer.write(type_name)
        serializer.write(value, buffer)

    @classmethod
    def read_existing(cls, fpath, intodic):
        '''TODO: Documentation
        '''
        lines = []
        with open(fpath, "r") as f:
            lines.extend(f.read().splitlines())
        buffer = GlobalDB.LineBuffer(lines)
        log_size = 0
        while buffer.has_next():
            op = buffer.read()
            assert op == "PUT"
            key = buffer.read()
            type_id = buffer.read()
            serializer = GlobalDB.serializers.get(type_id)
            value = serializer.read(buffer)
            intodic[key] = value
            log_size += 1
        return log_size

    @classmethod
    def write_put(cls, fname, key, value):
        '''TODO: Documentation
        '''
        type_name = str(type(value))
        serializer = cls.serializers.get(type_name)
        if not serializer:
            return  # no serializer for given value type
        if not os.path.exists(cls.get_storage_dir()):
            os.mkdir(cls.get_storage_dir())
        fpath = os.path.join(
            cls.get_storage_dir(),
            "{}.logdb.txt".format(fname)
        )
        buffer = GlobalDB.LineBuffer()
        cls.put_value(key, value, buffer)
        buffer.flush(fpath)

    @classmethod
    def read(cls, fname, intodic):
        '''TODO: Documentation
        '''
        fpath = os.path.join(
            cls.get_storage_dir(),
            "{}.logdb.txt".format(fname)
        )
        if os.path.exists(fpath):
            return cls.read_existing(fpath, intodic)
        else:
            return 0

    @classmethod
    def compress(cls, fname, content):
        '''TODO: Documentation
        '''
        buffer = GlobalDB.LineBuffer()
        for key in content:
            value = content[key]
            cls.put_value(key, value, buffer)
        fpath = os.path.join(
            cls.get_storage_dir(),
            "{}.logdb.txt".format(fname)
        )
        with open(fpath, "w") as f:
            f.writelines(buffer.buffer)

    def __init__(self, file_name):
        self.fname = file_name
        self.locked = {}
        self.content = {}

        filter(
            lambda a: a.__name__ == 'unload_nodes',
            bpy.app.handlers.game_post
        )
        remove_f = []
        for f in bpy.app.handlers.game_post:
            if f.__name__ == 'unload_nodes':
                remove_f.append(f)
        for f in remove_f:
            bpy.app.handlers.game_post.remove(f)
        bpy.app.handlers.game_post.append(unload_nodes)

        log_size = GlobalDB.read(self.fname, self.content)
        if log_size > (5 * len(self.content)):
            debug("Compressing sld {}".format(file_name))
            GlobalDB.compress(self.fname, self.content)

    def lock(self, item, event):
        self.locked[item] = event

    def unlock(self, item):
        self.locked.pop(item)

    def get(self, key, default_value=None):
        '''TODO: Documentation
        '''
        if not key:
            return default_value
        return self.content.get(key, default_value)

    def clear(self):
        '''TODO: Documentation
        '''
        self.content.clear()

    def put(self, key, value, persist=False):
        '''TODO: Documentation
        '''
        self.content[key] = value
        if persist:
            old_value = self.content.get(key)
            changed = old_value != value
            if changed:
                GlobalDB.write_put(self.fname, key, value)

    def check(self, key):
        '''TODO: Documentation
        '''
        valid = key in self.content
        return valid

    def pop(self, key, default=None):
        '''TODO: Documentation
        '''
        if not key:
            return default
        return self.content.pop(key, default)

    def remove(self, key):
        if key in self.content.keys():
            del self.content[key]

    def log(self):
        '''TODO: Documentation
        '''
        print(self.content)


def store(key, value, category='uplogic.default_globals'):
    initialize()
    values = GlobalDB.retrieve(category)
    values.put(key, value)


def retrieve(key, category='uplogic.default_globals', default=None):
    initialize()
    values = GlobalDB.retrieve(category)
    return values.get(key, default)


def initialize():
    if not GlobalDB.initialized:
        scene = logic.getCurrentScene()
        cats = getattr(
            bpy.data.scenes[scene.name],
            'nl_global_categories',
            None
        )
        if not cats:
            return

        msg = ''

        dat = {
            'STRING': 'string_val',
            'FLOAT': 'float_val',
            'INTEGER': 'int_val',
            'BOOLEAN': 'bool_val',
            'FILE_PATH': 'filepath_val'
        }

        for c in cats:
            db = GlobalDB.retrieve(c.name)
            msg += f' {c.name},'
            for v in c.content:
                val = getattr(v, dat.get(v.value_type, 'FLOAT'), 0)
                db.put(v.name, val, v.persistent)

        if msg:
            debug(f'Globals Initialized:{msg[:-1]}')
        GlobalDB.initialized = True


from .globaldb import GlobalDB
from .globaldb import retrieve  # noqa
from .globaldb import store  # noqa
from .file import read_file  # noqa
from .file import write_file  # noqa
from .file import load_file  # noqa
from .serializers import StringSerializer
from .serializers import FloatSerializer
from .serializers import IntegerSerializer
from .serializers import ListSerializer
from .serializers import VectorSerializer
from mathutils import Vector
from bge.types import KX_PythonComponent
from bge.types import KX_GameObject


GlobalDB.serializers[str(type(""))] = StringSerializer()
GlobalDB.serializers[str(type(1.0))] = FloatSerializer()
GlobalDB.serializers[str(type(10))] = IntegerSerializer()
GlobalDB.serializers[str(type([]))] = ListSerializer()
GlobalDB.serializers[str(type((0, 0, 0)))] = ListSerializer()
GlobalDB.serializers[str(type(Vector()))] = (
    VectorSerializer()
)


class GameProperty():

    def __init__(self, inst, name, default=None):
        
        def getPropComponent(self, attr_name=name):
                return self.object.get(attr_name)

        def setPropComponent(self, value, attr_name=name):
            self.object[attr_name] = value

        def getPropObject(self, attr_name=name):
            return self.get(attr_name)

        def setPropObject(self, value, attr_name=name):
            self[attr_name] = value

        if issubclass(inst.__class__, KX_PythonComponent):
            prop = property(getPropComponent, setPropComponent)
            if default:
                inst.object[name] = default
        elif issubclass(inst.__class__, KX_GameObject):
            prop = property(getPropObject, setPropObject)
            if default:
                inst[name] = default
        else:
            return
        setattr(inst.__class__, name, prop)




from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_INVALID
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULRunPython(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.module_name = None
        self.module_func = None
        self.arg = None
        self.val = None
        self.OUT = ULOutSocket(self, self.get_done)
        self.VAL = ULOutSocket(self, self.get_val)
        self._old_mod_name = None
        self._old_mod_fun = None
        self._module = None
        self._modfun = None

    def get_done(self):
        return self.done

    def get_val(self):
        return self.val

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        mname = self.get_input(self.module_name)
        mfun = self.get_input(self.module_func)
        if is_waiting(mname, mfun):
            return
        arg = self.get_input(self.arg)
        self._set_ready()
        if mname and (self._old_mod_name != mname):
            exec("import {}".format(mname))
            self._old_mod_name = mname
            self._module = eval(mname)
        if self._old_mod_fun != mfun:
            self._modfun = getattr(self._module, mfun)
            self._old_mod_fun = mfun
        if arg is STATUS_INVALID:
            self.val = self._modfun()
        else:
            self.val = self._modfun(arg)
        self.done = True



from uplogic.nodes import ULConditionNode
from uplogic.utils import STATUS_READY


class ULOnUpdate(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self._set_status(STATUS_READY)
        self._value = True

    def reset(self):
        self._value = True

    def evaluate(self):
        pass



from .cursorbehavior import ULCursorBehavior  # noqa
from .cursorvisibility import ULSetCursorVisibility  # noqa
from .endgame import ULEndGame  # noqa
from .gamepadvibration import ULGamepadVibration  # noqa
from .keylogger import ULKeyLogger  # noqa
from .loadblendfile import ULLoadBlendFile  # noqa
from .loadgame import ULLoadGame  # noqa
from .mouselook import ULMouseLook  # noqa
from .mousesetposition import ULSetMousePosition  # noqa
from .removeoverlaycollection import ULRemoveOverlayCollection  # noqa
from .restartgame import ULRestartGame  # noqa
from .runactuator import ULRunActuator  # noqa
from .runpython import ULRunPython  # noqa
from .savegame import ULSaveGame  # noqa
from .setactuatorvalue import ULSetActuatorValue  # noqa
from .setoverlaycollection import ULSetOverlayCollection  # noqa
from .setpyinstanceattr import ULSetPyInstanceAttr  # noqa
from .setsensorvalue import ULSetSensorValue  # noqa
from .startsound import ULStartSound  # noqa
from .startsound3d import ULStartSound3D  # noqa
from .startspeaker import ULStartSpeaker  # noqa
from .dispatchevent import ULDispatchEvent  # noqa
from .addobject import ULAddObject  # noqa
from .setproperty import ULSetProperty  # noqa
from .setmaterial import ULSetMaterial  # noqa
from .setmatnodevalue import ULSetMatNodeValue  # noqa
from .playsequence import ULPaySequence  # noqa
from .setnodesocket import ULSetNodeSocket  # noqa
from .setnodevalue import ULSetNodeValue  # noqa
from .toggleproperty import ULToggleProperty  # noqa
from .setmatnodesocket import ULSetMatNodeSocket  # noqa
from .copyproperty import ULCopyProperty  # noqa
from .modifyproperty import ULModifyProperty  # noqa
from .clampedmodifyproperty import ULClampedModifyProperty  # noqa
from .printvalue import ULPrintValue  # noqa
from .createvehicle import ULCreateVehicle  # noqa
from .vehicleapplyforce import ULVehicleApplyForce  # noqa
from .vehicleapplybraking import ULVehicleApplyBraking  # noqa
from .vehicleapplysteering import ULVehicleApplySteering  # noqa
from .vehiclesetattributes import ULVehicleSetAttributes  # noqa
from .setgameobjectattr import ULSetGameObjectAttribue  # noqa
from .sendmessage import ULSendMessage  # noqa
from .setvisibility import ULSetVisibility  # noqa
from .setvisibility import ULSetCollectionVisibility  # noqa
from .setcurvepoints import ULSetCurvePoints  # noqa
from .raycast import ULRaycast  # noqa
from .projectileraycast import ULProjectileRayCast  # noqa
from .mouseraycast import ULMouseRayCast  # noqa
from .cameraraycast import ULCameraRayCast  # noqa
from .setcamera import ULSetCamera  # noqa

from .addphysicsconstraint import ULAddPhysicsConstraint  # noqa
from .alignaxistovector import ULAlignAxisToVector  # noqa
from .appendlistitem import ULAppendListItem  # noqa
from .applyforce import ULApplyForce  # noqa
from .applyimpulse import ULApplyImpulse  # noqa
from .applymovement import ULApplyMovement  # noqa
from .applyrotation import ULApplyRotation  # noqa
from .applytorque import ULApplyTorque  # noqa
from .characterjump import ULCharacterJump  # noqa
from .clearvariables import ULClearVariables  # noqa
from .drawbox import ULDrawBox  # noqa
from .drawcube import ULDrawCube  # noqa
from .drawline import ULDrawLine  # noqa
from .editbone import ULEditBone  # noqa
from .endobject import ULEndObject  # noqa
from .executesubnetwork import ULExecuteSubNetwork  # noqa
from .followpath import ULFollowPath  # noqa
from .gamepadlook import ULGamepadLook  # noqa
from .getperformanceprofile import ULGetPerformanceProfile  # noqa
from .installsubnetwork import ULInstallSubNetwork  # noqa
from .listglobalvalues import ULListGlobalValues  # noqa
from .listvariables import ULListVariables  # noqa
from .makeuniquelight import ULMakeUniqueLight  # noqa
from .moveto import ULMoveTo  # noqa
from .movetowithnavmesh import ULMoveToWithNavmesh  # noqa
from .pausesound import ULPauseSound  # noqa
from .playaction import ULPlayAction  # noqa
from .popdictkey import ULPopDictKey  # noqa
from .removelistindex import ULRemoveListIndex  # noqa
from .removelistvalue import ULRemoveListValue  # noqa
from .removeparent import ULRemoveParent  # noqa
from .removephysicsconstraint import ULRemovePhysicsConstraint  # noqa
from .removevariable import ULRemoveVariable  # noqa
from .replacemesh import ULReplaceMesh  # noqa
from .resumesound import ULResumeSound  # noqa
# XXX. Rename Class to ULRotateTo
from .rotateto import ULActionRotateTo  # noqa
from .savevariable import ULSaveVariable  # noqa
from .savevariabledict import ULSaveVariableDict  # noqa
from .setactionframe import ULSetActionFrame  # noqa
from .setboneconstraintattr import ULSetBoneConstraintAttribute  # noqa
from .setboneconstraintinfluence import ULSetBoneConstraintInfluence  # noqa
from .setboneconstrainttarget import ULSetBoneConstraintTarget  # noqa
from .setboneposition import ULSetBonePosition  # noqa
from .setcamerafov import ULSetCameraFOV  # noqa
from .setcameraorthoscale import ULSetCameraOrthoScale  # noqa
from .setcharactergravity import ULSetCharacterGravity  # noqa
from .setcharacterjumpspeed import ULSetCharacterJumpSpeed  # noqa
from .setcharactermaxjumps import ULSetCharacterMaxJumps  # noqa
from .setcharactervelocity import ULSetCharacterVelocity  # noqa
from .setcharacterwalkdir import ULSetCharacterWalkDir  # noqa
from .setcollisiongroup import ULSetCollisionGroup  # noqa
from .setcollisionmask import ULSetCollisionMask  # noqa
from .setdictkey import ULSetDictKey  # noqa
from .setdynamics import ULSetDynamics  # noqa
from .seteeveeao import ULSetEeveeAO  # noqa
from .seteeveebloom import ULSetEeveeBloom  # noqa
from .seteeveesmaa import ULSetEeveeSMAA  # noqa
from .seteeveesmaaquality import ULSetEeveeSMAAQuality  # noqa
from .seteeveessr import ULSetEeveeSSR  # noqa
from .seteeveevolumetrics import ULSetEeveeVolumetrics  # noqa
from .setexposure import ULSetExposure  # noqa
from .setfullscreen import ULSetFullscreen  # noqa
from .setgamma import ULSetGamma  # noqa
from .setglobalvalue import ULSetGlobalValue  # noqa
from .setgravity import ULSetGravity  # noqa
from .setlightcolor import ULSetLightColor  # noqa
from .setlightenergy import ULSetLightEnergy  # noqa
from .setlightshadow import ULSetLightShadow  # noqa
from .setlistindex import ULSetListIndex  # noqa
from .setparent import ULSetParent  # noqa
from .setphysics import ULSetPhysics  # noqa
from .setprofile import ULSetProfile  # noqa
from .setresolution import ULSetResolution  # noqa
from .setrigidbody import ULSetRigidBody  # noqa
from .settimescale import ULSetTimeScale  # noqa
from .setvsync import ULSetVSync  # noqa
from .showframerate import ULShowFramerate  # noqa
from .startsubnetwork import ULStartSubNetwork  # noqa
from .stopaction import ULStopAction  # noqa
from .stopallsounds import ULStopAllSounds  # noqa
from .stopsound import ULStopSound  # noqa
from .stopsubnetwork import ULStopSubNetwork  # noqa
from .translate import ULTranslate  # noqa
from .addfilter import ULAddFilter  # noqa
from .removefilter import ULRemoveFilter  # noqa
from .togglefilter import ULToggleFilter  # noqa
from .setfilterstate import ULSetFilterState  # noqa
from .setscene import ULSetScene  # noqa
from .loadscene import ULLoadScene  # noqa
from .loadfilecontent import ULLoadFileContent  # noqa
from .slowfollow import ULSlowFollow  # noqa


from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import lerp
from uplogic.utils import is_invalid


class ULInterpolate(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.a = None
        self.b = None
        self.fac = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        socket = self.get_output('val')
        if socket is None:
            a = self.get_input(self.a)
            b = self.get_input(self.b)
            fac = self.get_input(self.fac)
            if is_invalid(a, b, fac):
                return STATUS_WAITING
            return self.set_output('val', lerp(a, b, fac))
        return socket

    def evaluate(self):
        self._set_ready()



from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.utils import is_invalid
from uplogic.utils import not_met
from uplogic.utils import draw_cube


class ULDrawCube(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.color = None
        self.origin = None
        self.width = None
        self.use_volume_origin = False

    def evaluate(self):
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        origin = self.get_input(self.origin)
        width = self.get_input(self.width)
        color = self.get_input(self.color)
        if is_invalid(origin, width, color):
            return
        origin = origin.copy()
        self._set_ready()
        if self.use_volume_origin:
            offset = width * 0.5
            origin -= Vector((offset, offset, offset))
        draw_cube(origin, width, color)
        self._set_value(True)



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting


class ULListFromItems(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.items = None
        self.list: list = None
        self.LIST = ULOutSocket(self, self.get_list)

    def get_list(self):
        socket = self.get_output('list')
        if socket is None:
            self.list = self.get_input(self.items)
            return self.set_output('list', [self.get_input(item) for item in self.list])
        return socket

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULGetPerformanceProfile(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.network = None
        self.print_profile = False
        self.check_evaluated_cells = False
        self.check_average_cells_per_sec = False
        self.check_cells_per_tick = False
        self.done = None
        self.data = ''
        self.OUT = ULOutSocket(self, self.get_done)
        self.DATA = ULOutSocket(self, self.get_data)

    def get_done(self):
        return self.done

    def get_data(self):
        return self.data

    def setup(self, network):
        self.network = network

    def evaluate(self):
        self.done = False
        self.data = '----------------------------------Start Profile\n'
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            return
        print_profile = self.get_input(
            self.print_profile
        )
        check_evaluated_cells = self.get_input(
            self.check_evaluated_cells
        )
        check_average_cells_per_sec = self.get_input(
            self.check_average_cells_per_sec
        )
        check_cells_per_tick = self.get_input(
            self.check_cells_per_tick
        )
        if is_waiting(
            print_profile,
            check_evaluated_cells,
            check_average_cells_per_sec,
            check_cells_per_tick
        ):
            self._set_ready()
            return
        self._set_ready()
        if check_evaluated_cells:
            self.data += 'Evaluated Nodes:\t{}\n'.format(
                self.network.evaluated_cells
            )
        if check_average_cells_per_sec:
            self.data += 'Nodes per Sec (avg):\t{}\n'.format(
                self.network.evaluated_cells / self.network.timeline
            )
        if check_cells_per_tick:
            self.data += 'Nodes per Tick:\t{}\n'.format(
                len(self.network._cells)
            )
        self.data += '----------------------------------End Profile'
        if print_profile:
            print(self.data)
        self.done = True



from bge import logic
import bpy
# from uplogic.events import schedule


def set_scene(scene: str or bpy.types.Scene) -> None:
    logic.getCurrentScene().replace(scene)


class FileLoader():

    def __init__(self, start=True):
        self.status = 0.0
        self.item = ''
        self.data = 'textures'
        self.meshes = [mesh for mesh in bpy.data.meshes]
        self.materials = [m for m in bpy.data.materials if not m.is_grease_pencil]
        self.images = [img for img in bpy.data.images]
        self.datasize = len(self.meshes) + len(self.materials) + len(self.images)
        self.finished = False
        if start:
            self.start()
    
    def start(self):
        self.create_object()
        logic.getCurrentScene().pre_draw.append(self.load_next)

    def create_object(self):
        self.bmesh = bmesh = bpy.data.meshes.new('ContentLoader')
        self.temp_map = temp_mat = bpy.data.materials.new('ContentLoader')
        temp_mat.use_nodes = True
        self.tex_image = tex_image = temp_mat.node_tree.nodes.new('ShaderNodeTexImage')
        temp_mat.node_tree.links.new(tex_image.outputs[0], temp_mat.node_tree.nodes['Principled BSDF'].inputs[0])
        self.bobj = bobj = bpy.data.objects.new('ContentLoader', bmesh)
        bpy.context.collection.objects.link(bobj)
        bmesh.materials.append(temp_mat)

        self.object = logic.getCurrentScene().convertBlenderObject(bobj)
        self.object.worldScale = (0.000001, 0.000001, 0.0000001)

    def load_next(self):
        cam = logic.getCurrentScene().active_camera
        self.object.worldPosition = cam.worldPosition - cam.getAxisVect((0, 0, 1)) * 100
        if self.images:
            self.tex_image.image = self.images.pop()
            self.status += 1 / self.datasize
            self.item = self.tex_image.image.name
            return
        if self.materials:
            mat = self.materials.pop()
            self.object.blenderObject.material_slots[0].material = mat
            self.status += 1 / self.datasize
            self.data = 'shaders'
            self.item = mat.name
            return
        if self.meshes:
            self.bobj.data = self.meshes.pop()
            self.status += 1 / self.datasize
            self.data = 'objects'
            self.item = self.bobj.data.name
            return
        logic.getCurrentScene().pre_draw.remove(self.load_next)
        self.object.endObject()
        bpy.data.materials.remove(self.temp_map)
        bpy.data.meshes.remove(self.bmesh)
        bpy.data.objects.remove(self.bobj)
        self.finished = True
        self.on_finish()

    def on_finish(self):
        pass


class SceneLoader():

    def __init__(self, scene: str, start=True):
        if isinstance(scene, str):
            scene = bpy.data.scenes.get(scene)
        if not isinstance(scene, bpy.types.Scene):
            print(f'SceneLoader: Scene {scene} not found!')
            return

        self.scene = scene
        self.status = 0.0
        self.item = ''
        self.data = 'textures'
        self.meshes = []
        self.materials = []
        self.images = [img for img in bpy.data.images]
        self.finished = False
        self.fetch_data()
        self.datasize = len(self.meshes) + len(self.materials) + len(self.images)
        if start:
            self.start()
    
    def fetch_data(self):
        for bobj in self.scene.objects:
            if isinstance(bobj.data, bpy.types.Mesh) and bobj.data not in self.meshes:
                self.meshes.append(bobj.data)
            for slot in bobj.material_slots:
                if slot.material not in self.materials:
                    self.materials.append(slot.material)

    def start(self):
        self.create_object()
        logic.getCurrentScene().pre_draw.append(self.load_next)

    def create_object(self):
        self.bmesh = bmesh = bpy.data.meshes.new('ContentLoader')
        self.temp_map = temp_mat = bpy.data.materials.new('ContentLoader')
        temp_mat.use_nodes = True
        self.tex_image = tex_image = temp_mat.node_tree.nodes.new('ShaderNodeTexImage')
        temp_mat.node_tree.links.new(tex_image.outputs[0], temp_mat.node_tree.nodes['Principled BSDF'].inputs[0])
        self.bobj = bobj = bpy.data.objects.new('ContentLoader', bmesh)
        bpy.context.collection.objects.link(bobj)
        bmesh.materials.append(temp_mat)

        self.object = logic.getCurrentScene().convertBlenderObject(bobj)
        self.object.worldScale = (0.000001, 0.000001, 0.0000001)

    def load_next(self):
        cam = logic.getCurrentScene().active_camera
        self.object.worldPosition = cam.worldPosition - cam.getAxisVect((0, 0, 1)) * 100
        if self.images:
            self.tex_image.image = self.images.pop()
            self.status += 1 / self.datasize
            self.item = self.tex_image.image.name
            return
        if self.materials:
            mat = self.materials.pop()
            self.object.blenderObject.material_slots[0].material = mat
            self.status += 1 / self.datasize
            self.data = 'shaders'
            self.item = mat.name
            return
        if self.meshes:
            self.bobj.data = self.meshes.pop()
            self.status += 1 / self.datasize
            self.data = 'objects'
            self.item = self.bobj.data.name
            return
        logic.getCurrentScene().pre_draw.remove(self.load_next)
        self.object.endObject()
        bpy.data.materials.remove(self.temp_map)
        bpy.data.meshes.remove(self.bmesh)
        bpy.data.objects.remove(self.bobj)
        self.finished = True
        self.on_finish()

    def on_finish(self):
        pass
